/*

    this source file is a part of 
    lp_parse - a preprocessor for MILP models
    Copyright (C) 1996  Ari Jolma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    for more information please contact 
    Ari Jolma
    Helsinki University of Technology
    Laboratory of Water Resources Engineering
    Building W
    02150 Espoo, Finland
    E-mail: Ari.Jolma@hut.fi

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "util.h"

FILE *mystderr = stderr;
char *pn = NULL;

/* --------------------------------------------------------------------------*/

void *xmalloc(size_t size) {
  void *tmp;
  tmp = malloc(size);
  if ((tmp == NULL) && (size > 0)) 
    fprintf(mystderr, "Out of memory (%i)\n", size);
  return tmp;
}

void *xrealloc(void *ptr, size_t size) {
  void *tmp;
  tmp = realloc(ptr, size);
  if ((tmp == NULL) && (size > 0)) 
    fprintf(mystderr, "Out of memory (%i)\n", size);
  return tmp;
}

void *xcalloc(size_t nmemb, size_t size) {
  void *tmp;
  tmp = calloc(nmemb, size);
  if ((tmp == NULL) && (nmemb*size > 0)) 
    fprintf(mystderr, "Out of memory (%li)\n", (long)nmemb*size);
  return tmp;
}

/* --------------------------------------------------------------------------*/

int IsSubClass(CNR Super, CNR Sub) {
  while (Super) {
#ifdef DEBUG
    /*printf("%X\t%X\n", Super, Sub);*/
#endif
    if ((Super % CLASS_NUMBER_BASE) != (Sub % CLASS_NUMBER_BASE)) return FALSE;
    Super /= CLASS_NUMBER_BASE;
    Sub /= CLASS_NUMBER_BASE;
  }
  return TRUE;
}

CNR getcnr(char *cn, char **cns, CNR *cnrs, int *n) {
  do {
    (*n)--;
    if (!strcmp(cn, cns[*n])) 
      return cnrs[*n];
  } while (*n > 0);
  return 0;
}

char *getcn(CNR cnr, char **cns, CNR *cnrs, int *n) {
  do {
    (*n)--;
    if (cnr == cnrs[*n])
      return cns[*n];
  } while (*n > 0);
  return NULL;
}

/* --------------------------------------------------------------------------*/

char *xstrdup(const char *c) {
  char *t;
  if (!c) return NULL;
  t = (char *) xmalloc(strlen(c)+1);
  if (t == NULL) return NULL;
  return strcpy(t, c);
}

/* --------------------------------------------------------------------------*/

char *xgetline(XFILE *f, char *buf, int buf_size) {
  register int i=0;
  if (f->f == NULL) {
    if (f->cln >= f->lines) return NULL;
    strncpy(buf, f->data[f->cln], buf_size);
    (f->cln)++;
    return buf;
  }
  if (feof(f->f)) return NULL;
  while (TRUE) {
    buf[i] = (char)fgetc(f->f);
    if (feof(f->f) || (buf[i] == 4) || (buf[i] == '\n')) {
      buf[i] = '\r';
      buf[i+1] = '\0';
      (f->cln)++;
      return buf;
    }
    if (i >= (buf_size-2)) {
      fprintf(mystderr, "%s: %s: %i: too long input line\n", 
	      pn, filename(f->name), f->cln);
      buf[i] = '\0';
      (f->cln)++;
      return buf;
    }
    ++i;
  }
}

char *getextline(XFILE *f, char *buf, int buf_size) {
  int ext;
  char line[MAX_LINE_LEN], *s;
  buf[0] = '\0';
  do {
    s = xgetline(f, line, MAX_LINE_LEN);
    if (s == NULL) return NULL;
    while (*s) s++;    
    do
      s--;
    while (((*s == ' ') || (*s == '\t') || (*s == '\r') || (*s == '\n')) && 
	   (s != line));
    ext = strchr(LINE_CONTINUATION_CHARS, *s) != NULL;
    if (!ext) s++;
    *s = '\0';
    strncat(buf, line, buf_size - strlen(buf)-1);
  } while (ext);
  return buf;
}

/* --------------------------------------------------------------------------*/

char *filename(char *path) {
  char *tmp = path;
  if (path == NULL) return NULL;
  while (*path) path++;
  while ((path != tmp) && ((*path != '/') || (*path != '\\'))) path--;
  if ((*path == '/') || (*path == '\\')) path++;
  return path;
}

XFILE *new_XFILE(char *name) {
  XFILE *xfile;
  xfile  = (XFILE *) xmalloc(sizeof(XFILE));
  if (xfile == NULL) return NULL;
  if (name == NULL) 
    xfile->name = NULL; 
  else {
    xfile->name = strdup(name);
    if (xfile->name == NULL) return NULL;
  }
  xfile->cln = 0;
  xfile->maxlines = 0;
  xfile->lines = 0;
  xfile->data = NULL;
  return xfile;
}

XFILE *xstdin() {
  XFILE *xfile = new_XFILE(NULL);  
  if (xfile == NULL) return NULL;
  xfile->f = stdin;
  return xfile;
}

XFILE *xstdout() {
  XFILE *xfile = new_XFILE(NULL);  
  if (xfile == NULL) return NULL;
  xfile->f = stdout;
  return xfile;
}

XFILE *open_file(char *name, char *flags) {
  XFILE *xfile = new_XFILE(name);  
  char line[MAX_LINE_LEN], *s;
  if (xfile == NULL) return NULL;
  xfile->f = fopen(name, flags);
  if (!(xfile->f)) {
    if (flags[0] == 'r')
      fprintf(mystderr, "%s: %s: file not found\n", pn, name);
    else
      fprintf(mystderr, "%s: %s: file could not be opened\n", pn, name);
    return NULL;
  } 
  if (flags[0] == 'r') {
    xfile->maxlines = 10000;
    xfile->data  = (char **) xmalloc(xfile->maxlines * sizeof(char **));
    if (xfile->data == NULL) return NULL;

#ifdef DEBUG
    /*printf("reading %s\n", name);*/
#endif

    while ((s = xgetline(xfile, line, MAX_LINE_LEN)) != NULL) {

#ifdef DEBUG
      /*printf("%s\n", line);*/
#endif

      if (xfile->lines > xfile->maxlines) {
	fprintf(mystderr, "%s:%s too big file, sorry\n", pn, name);
	return NULL;
      }
      xfile->data[xfile->cln - 1] = xstrdup(line);
      if (xfile->data[xfile->cln - 1] == NULL) return NULL;
    }
    fclose(xfile->f);
    xfile->lines = xfile->cln;
    xfile->cln = 0;
    xfile->f = NULL;    
  }
  return xfile;
}

char *xfn(XFILE *xf) {
  if (xf == NULL) return NULL;
  return xf->name;
}
